@@ -46,8 +46,11 @@
 #include <assert.h>
 #include <pthread.h>
 
+#define NBDKIT_API_VERSION 2
+
 #include <nbdkit-plugin.h>
 #include "protocol.h"
+#include "byte-swapping.h"
 
 static char *sockname = NULL;
 static char *export = NULL;
@@ -98,7 +101,7 @@ nbd_config_complete (void)
     nbdkit_error ("you must supply the socket=<SOCKNAME> parameter after the plugin name on the command line");
     return -1;
   }
-  if (strlen (sockname) >= sizeof sock.sun_path) {
+  if (strlen (sockname) > sizeof sock.sun_path) {
     nbdkit_error ("socket file name too large");
     return -1;
   }
@@ -463,9 +466,11 @@ nbd_open (int readonly)
   h->fd = socket (AF_UNIX, SOCK_STREAM, 0);
   if (h->fd < 0) {
     nbdkit_error ("socket: %m");
+    free (h);
     return NULL;
   }
-  strncpy (sock.sun_path, sockname, sizeof (sock.sun_path));
+  /* We already validated length during nbd_config_complete */
+  memcpy (sock.sun_path, sockname, strlen (sockname));
   if (connect (h->fd, (const struct sockaddr *) &sock, sizeof sock) < 0) {
     nbdkit_error ("connect: %m");
     goto err;
@@ -556,6 +561,7 @@ nbd_open (int readonly)
 
  err:
   close (h->fd);
+  free (h);
   return NULL;
 }
 
@@ -565,8 +571,10 @@ nbd_close (void *handle)
 {
   struct handle *h = handle;
 
-  if (!h->dead)
+  if (!h->dead) {
     nbd_request_raw (h, 0, NBD_CMD_DISC, 0, 0, 0, NULL);
+    shutdown (h->fd, SHUT_WR);
+  }
   close (h->fd);
   if ((errno = pthread_join (h->reader, NULL)))
     nbdkit_debug ("failed to join reader thread: %m");
@@ -616,68 +624,89 @@ nbd_can_trim (void *handle)
   return h->flags & NBD_FLAG_SEND_TRIM;
 }
 
+static int
+nbd_can_zero (void *handle)
+{
+  struct handle *h = handle;
+
+  return h->flags & NBD_FLAG_SEND_WRITE_ZEROES;
+}
+
+static int
+nbd_can_fua (void *handle)
+{
+  struct handle *h = handle;
+
+  return h->flags & NBD_FLAG_SEND_FUA ? NBDKIT_FUA_NATIVE : NBDKIT_FUA_NONE;
+}
+
 /* Read data from the file. */
 static int
-nbd_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
+nbd_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
+           uint32_t flags)
 {
   struct handle *h = handle;
   int c;
 
-  /* TODO Auto-fragment this if the client has a larger max transfer
-     limit than the server */
+  assert (!flags);
   c = nbd_request_full (h, 0, NBD_CMD_READ, offset, count, NULL, buf);
   return c < 0 ? c : nbd_reply (h, c);
 }
 
 /* Write data to the file. */
 static int
-nbd_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset)
+nbd_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
+            uint32_t flags)
 {
   struct handle *h = handle;
   int c;
 
-  /* TODO Auto-fragment this if the client has a larger max transfer
-     limit than the server */
-  c = nbd_request_full (h, 0, NBD_CMD_WRITE, offset, count, buf, NULL);
+  assert (!(flags & ~NBDKIT_FLAG_FUA));
+  c = nbd_request_full (h, flags & NBDKIT_FLAG_FUA ? NBD_CMD_FLAG_FUA : 0,
+                        NBD_CMD_WRITE, offset, count, buf, NULL);
   return c < 0 ? c : nbd_reply (h, c);
 }
 
 /* Write zeroes to the file. */
 static int
-nbd_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+nbd_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
   struct handle *h = handle;
   int c;
+  int f = 0;
 
-  if (!(h->flags & NBD_FLAG_SEND_WRITE_ZEROES)) {
-    /* Trigger a fall back to regular writing */
-    errno = EOPNOTSUPP;
-    return -1;
-  }
+  assert (!(flags & ~(NBDKIT_FLAG_FUA | NBDKIT_FLAG_MAY_TRIM)));
+  assert (h->flags & NBD_FLAG_SEND_WRITE_ZEROES);
 
-  c = nbd_request (h, may_trim ? 0 : NBD_CMD_FLAG_NO_HOLE,
-                   NBD_CMD_WRITE_ZEROES, offset, count);
+  if (!(flags & NBDKIT_FLAG_MAY_TRIM))
+    f |= NBD_CMD_FLAG_NO_HOLE;
+  if (flags & NBDKIT_FLAG_FUA)
+    f |= NBD_CMD_FLAG_FUA;
+  c = nbd_request (h, f, NBD_CMD_WRITE_ZEROES, offset, count);
   return c < 0 ? c : nbd_reply (h, c);
 }
 
 /* Trim a portion of the file. */
 static int
-nbd_trim (void *handle, uint32_t count, uint64_t offset)
+nbd_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
   struct handle *h = handle;
   int c;
 
-  c = nbd_request (h, 0, NBD_CMD_TRIM, offset, count);
+  assert (!(flags & ~NBDKIT_FLAG_FUA));
+  c = nbd_request (h, flags & NBDKIT_FLAG_FUA ? NBD_CMD_FLAG_FUA : 0,
+                   NBD_CMD_TRIM, offset, count);
   return c < 0 ? c : nbd_reply (h, c);
 }
 
 /* Flush the file to disk. */
 static int
-nbd_flush (void *handle)
+nbd_flush (void *handle, uint32_t flags)
 {
   struct handle *h = handle;
   int c;
 
+  assert (!flags);
   c = nbd_request (h, 0, NBD_CMD_FLUSH, 0, 0);
   return c < 0 ? c : nbd_reply (h, c);
 }
@@ -697,6 +726,8 @@ static struct nbdkit_plugin plugin = {
   .can_flush          = nbd_can_flush,
   .is_rotational      = nbd_is_rotational,
   .can_trim           = nbd_can_trim,
+  .can_zero           = nbd_can_zero,
+  .can_fua            = nbd_can_fua,
   .pread              = nbd_pread,
   .pwrite             = nbd_pwrite,
   .zero               = nbd_zero,
