@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2014 Red Hat Inc.
+ * Copyright (C) 2013-2018 Red Hat Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -69,7 +69,7 @@ set_error (PyObject *self, PyObject *args)
 {
   int err;
 
-  if (!PyArg_ParseTuple(args, "i", &err))
+  if (!PyArg_ParseTuple (args, "i", &err))
     return NULL;
   nbdkit_set_error (err);
   last_error = err;
@@ -94,6 +94,11 @@ callback_defined (const char *name, PyObject **obj_rtn)
   obj = PyObject_GetAttrString (module, name);
   if (!obj)
     return 0;
+  if (!PyCallable_Check (obj)) {
+    nbdkit_debug ("object %s isn't callable", name);
+    Py_DECREF (obj);
+    return 0;
+  }
 
   if (obj_rtn != NULL)
     *obj_rtn = obj;
@@ -103,13 +108,113 @@ callback_defined (const char *name, PyObject **obj_rtn)
   return 1;
 }
 
+/* Convert bytes/str/unicode into a string.  Caller must free. */
+static char *
+python_to_string (PyObject *str)
+{
+  if (str) {
+#ifdef HAVE_PYUNICODE_ASUTF8
+    if (PyUnicode_Check (str))
+      return strdup (PyUnicode_AsUTF8 (str));
+    else
+#endif
+#ifdef HAVE_PYSTRING_ASSTRING
+    if (PyString_Check (str))
+      return strdup (PyString_AsString (str));
+    else
+#endif
+    if (PyBytes_Check (str))
+      return strdup (PyBytes_AS_STRING (str));
+  }
+  return NULL;
+}
+
+/* This is the fallback in case we cannot get the full traceback. */
+static void
+print_python_error (const char *callback, PyObject *error)
+{
+  PyObject *error_str;
+  char *error_cstr = NULL;
+
+  error_str = PyObject_Str (error);
+  error_cstr = python_to_string (error_str);
+  nbdkit_error ("%s: %s: error: %s",
+                script, callback,
+                error_cstr ? error_cstr : "<unknown>");
+  Py_DECREF (error_str);
+  free (error_cstr);
+}
+
+/* Convert the Python traceback to a string and call nbdkit_error.
+ * https://stackoverflow.com/a/15907460/7126113
+ */
+static int
+print_python_traceback (const char *callback,
+                        PyObject *type, PyObject *error, PyObject *traceback)
+{
+  PyObject *module_name, *traceback_module, *format_exception_fn, *rv,
+    *traceback_str;
+  char *traceback_cstr;
+
+#ifdef HAVE_PYSTRING_FROMSTRING
+  module_name = PyString_FromString ("traceback");
+#else
+  module_name = PyUnicode_FromString ("traceback");
+#endif
+  traceback_module = PyImport_Import (module_name);
+  Py_DECREF (module_name);
+
+  /* couldn't 'import traceback' */
+  if (traceback_module == NULL)
+    return -1;
+
+  format_exception_fn = PyObject_GetAttrString (traceback_module,
+                                                "format_exception");
+  if (format_exception_fn == NULL)
+    return -1;
+  if (!PyCallable_Check (format_exception_fn))
+    return -1;
+
+  rv = PyObject_CallFunctionObjArgs (format_exception_fn,
+                                     type, error, traceback, NULL);
+  traceback_str = PyObject_Str (rv);
+  Py_DECREF (rv);
+  traceback_cstr = python_to_string (traceback_str);
+  if (traceback_cstr == NULL) {
+    Py_DECREF (traceback_str);
+    return -1;
+  }
+
+  nbdkit_error ("%s: %s: error: %s",
+                script, callback,
+                traceback_cstr);
+  Py_DECREF (traceback_str);
+  free (traceback_cstr);
+
+  /* This means we succeeded in calling nbdkit_error. */
+  return 0;
+}
+
 static int
 check_python_failure (const char *callback)
 {
   if (PyErr_Occurred ()) {
-    nbdkit_error ("%s: callback failed: %s", script, callback);
-    /* How to turn this into a string? XXX */
-    PyErr_Print ();
+    PyObject *type, *error, *traceback;
+
+    PyErr_Fetch (&type, &error, &traceback);
+    PyErr_NormalizeException (&type, &error, &traceback);
+
+    /* Try to print the full traceback. */
+    if (print_python_traceback (callback, type, error, traceback) == -1) {
+      /* Couldn't do that, so fall back to converting the Python error
+       * to a string.
+       */
+      print_python_error (callback, error);
+    }
+
+    /* In all cases this returns -1 to indicate that a Python error
+     * occurred.
+     */
     return -1;
   }
   return 0;
@@ -136,15 +241,14 @@ py_load (void)
 #if PY_MAJOR_VERSION >= 3
   PyModule_Create (&moduledef);
 #else
-  Py_InitModule("nbdkit", NbdkitMethods);
+  Py_InitModule ("nbdkit", NbdkitMethods);
 #endif
 }
 
 static void
 py_unload (void)
 {
-  if (module)
-    Py_DECREF (module);
+  Py_XDECREF (module);
 
   Py_Finalize ();
 }
@@ -155,7 +259,7 @@ py_dump_plugin (void)
   PyObject *fn;
   PyObject *r;
 
-  if (callback_defined ("dump_plugin", &fn)) {
+  if (script && callback_defined ("dump_plugin", &fn)) {
     PyErr_Clear ();
 
     r = PyObject_CallObject (fn, NULL);
@@ -170,7 +274,6 @@ py_config (const char *key, const char *value)
   FILE *fp;
   PyObject *modname;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (!script) {
@@ -219,17 +322,8 @@ py_config (const char *key, const char *value)
     /* Other parameters are passed to the Python .config callback. */
     PyErr_Clear ();
 
-    args = PyTuple_New (2);
-#ifdef HAVE_PYSTRING_FROMSTRING
-    PyTuple_SetItem (args, 0, PyString_FromString (key));
-    PyTuple_SetItem (args, 1, PyString_FromString (value));
-#else
-    PyTuple_SetItem (args, 0, PyUnicode_FromString (key));
-    PyTuple_SetItem (args, 1, PyUnicode_FromString (value));
-#endif
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunction (fn, "ss", key, value);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("config") == -1)
       return -1;
     Py_DECREF (r);
@@ -267,7 +361,6 @@ static void *
 py_open (int readonly)
 {
   PyObject *fn;
-  PyObject *args;
   PyObject *handle;
 
   if (!callback_defined ("open", &fn)) {
@@ -277,11 +370,9 @@ py_open (int readonly)
 
   PyErr_Clear ();
 
-  args = PyTuple_New (1);
-  PyTuple_SetItem (args, 0, PyBool_FromLong (readonly));
-  handle = PyObject_CallObject (fn, args);
+  handle = PyObject_CallFunctionObjArgs (fn, readonly ? Py_True : Py_False,
+                                         NULL);
   Py_DECREF (fn);
-  Py_DECREF (args);
   if (check_python_failure ("open") == -1)
     return NULL;
 
@@ -293,21 +384,15 @@ py_close (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (callback_defined ("close", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     check_python_failure ("close");
-    if (r)
-      Py_DECREF (r);
+    Py_XDECREF (r);
   }
 
   Py_DECREF (obj);
@@ -318,7 +403,6 @@ py_get_size (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
   int64_t ret;
 
@@ -329,12 +413,8 @@ py_get_size (void *handle)
 
   PyErr_Clear ();
 
-  args = PyTuple_New (1);
-  Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-  PyTuple_SetItem (args, 0, obj);
-  r = PyObject_CallObject (fn, args);
+  r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
   Py_DECREF (fn);
-  Py_DECREF (args);
   if (check_python_failure ("get_size") == -1)
     return -1;
 
@@ -352,7 +432,6 @@ py_pread (void *handle, void *buf,
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (!callback_defined ("pread", &fn)) {
@@ -362,14 +441,8 @@ py_pread (void *handle, void *buf,
 
   PyErr_Clear ();
 
-  args = PyTuple_New (3);
-  Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-  PyTuple_SetItem (args, 0, obj);
-  PyTuple_SetItem (args, 1, PyLong_FromLong (count));
-  PyTuple_SetItem (args, 2, PyLong_FromUnsignedLongLong (offset));
-  r = PyObject_CallObject (fn, args);
+  r = PyObject_CallFunction (fn, "OiL", obj, count, offset, NULL);
   Py_DECREF (fn);
-  Py_DECREF (args);
   if (check_python_failure ("pread") == -1)
     return -1;
 
@@ -398,20 +471,15 @@ py_pwrite (void *handle, const void *buf,
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (callback_defined ("pwrite", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (3);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    PyTuple_SetItem (args, 1, PyByteArray_FromStringAndSize (buf, count));
-    PyTuple_SetItem (args, 2, PyLong_FromUnsignedLongLong (offset));
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunction (fn, "ONL", obj,
+                               PyByteArray_FromStringAndSize (buf, count),
+                               offset, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("pwrite") == -1)
       return -1;
     Py_DECREF (r);
@@ -429,18 +497,13 @@ py_flush (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (callback_defined ("flush", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("flush") == -1)
       return -1;
     Py_DECREF (r);
@@ -458,20 +521,13 @@ py_trim (void *handle, uint32_t count, uint64_t offset)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (callback_defined ("trim", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (3);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    PyTuple_SetItem (args, 1, PyLong_FromLong (count));
-    PyTuple_SetItem (args, 2, PyLong_FromUnsignedLongLong (offset));
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunction (fn, "OiL", obj, count, offset, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("trim") == -1)
       return -1;
     Py_DECREF (r);
@@ -510,8 +566,7 @@ py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
          gracefully fall back, and to accomodate both a normal return
          and an exception. */
       nbdkit_debug ("zero requested falling back to pwrite");
-      if (r)
-        Py_DECREF (r);
+      Py_XDECREF (r);
       PyErr_Clear ();
       return -1;
     }
@@ -531,19 +586,14 @@ py_can_write (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
   int ret;
 
   if (callback_defined ("can_write", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("can_write") == -1)
       return -1;
     ret = r == Py_True;
@@ -564,19 +614,14 @@ py_can_flush (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
   int ret;
 
   if (callback_defined ("can_flush", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("can_flush") == -1)
       return -1;
     ret = r == Py_True;
@@ -597,19 +642,14 @@ py_is_rotational (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
   int ret;
 
   if (callback_defined ("is_rotational", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("is_rotational") == -1)
       return -1;
     ret = r == Py_True;
@@ -625,19 +665,14 @@ py_can_trim (void *handle)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
   int ret;
 
   if (callback_defined ("can_trim", &fn)) {
     PyErr_Clear ();
 
-    args = PyTuple_New (1);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    Py_DECREF (args);
     if (check_python_failure ("can_trim") == -1)
       return -1;
     ret = r == Py_True;
@@ -687,4 +722,4 @@ static struct nbdkit_plugin plugin = {
   .zero              = py_zero,
 };
 
-NBDKIT_REGISTER_PLUGIN(plugin)
+NBDKIT_REGISTER_PLUGIN (plugin)
